#! /usr/bin/env python3

import os
import re
from sys        import exit, path
path.insert(0,"/opt/ark-ping-beacon/lib")
from re     	import search, MULTILINE
from subprocess import call
from random		import random
from time		import sleep, time
from json		import loads
from ping		import echo, verbose

#ark_dir = os.path.realpath(__file__)	# /usr/local/ark/bin/ark-ping-beacon.py
#ark_dir = os.path.dirname(ark_dir)		# /usr/local/ark/bin
#ark_dir = os.path.dirname(ark_dir)		# /usr/local/ark
ark_dir = "/usr/local/ark"
ark_ping_apps = "/opt/ark-ping-beacon/etc/ark-ping-apps.json"
ark_activity_config = "/etc/ark/activity-config"
ark_monitor_config = "/etc/ark/monitor-config"
activity_folder = "/usr/local/ark/activity"

def vnstat_rate():
	uname = os.uname()[0]

	uu = 'd@' if uname == 'Linux' else 'r'
	tt = os.popen('date -'+uu+'$((`date +%s`-86400)) +%m/%d/%Y').read().split()[0]
	uu = os.popen('date -'+uu+'$((`date +%s`-86400)) +%Y-%m-%d').read().split()[0]
	uu = '(?:'+tt+'|'+uu+')'

	rate = '0'
	if uname == 'Linux':
		m = os.popen('vnstat -d 2> /dev/null').read()
		if m:
			m = search('.*%s.*?([\d\.]+)\skbit/s'%uu,m)
			if m:
				rate = m.group(1)

	return tt[6:]+'/'+tt[0:2]+'/'+tt[3:5]+'='+rate

def interface_name():
	return os.popen('ip link show').read().split('\n')[0:-1][2].split()[1][0:-1]

def test_ping6( ipv6 ):
	ok = call(['ping6','-c','3','-W','3',ipv6],stdout=open('/dev/null','w'),stderr=open('/dev/null','w'))
	return '%s'%ok	# 0: success; 1: packet dropped; 2: other error

def network_rpi():
	(ipv4,ipv6_native,ipv6_teredo,rx,tx,ping6_native,ping6_teredo) = ('','','','0','0','2','2')

	for ip in os.popen('hostname -I').read().split('\n')[0].split():
		m = search('\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}',ip)
		if m:
			ipv4 = ip
		elif ip.startswith('2001:0:c0ac:e290'):
			ipv6_teredo = ip
		else:
			ipv6_native = ip

	# Works for RPi and Ubuntu
	# ark@zrh2-ch:/usr/local/ark/bin$ ip link show
	# 1: lo: <LOOPBACK,UP,LOWER_UP> mtu 65536 qdisc noqueue state UNKNOWN mode DEFAULT group default qlen 1000
	#    link/loopback 00:00:00:00:00:00 brd 00:00:00:00:00:00
	# 2: enp0s25: <BROADCAST,MULTICAST,UP,LOWER_UP> mtu 1500 qdisc fq_codel state UP mode DEFAULT group default qlen 1000
	#    link/ether 00:1c:c0:b2:ae:69 brd ff:ff:ff:ff:ff:ff

	interface = interface_name()

	ff = "/sys/class/net/%s/statistics/rx_bytes"%interface
	rx = open(ff).read().split('\n')[0] if os.path.exists(ff) else "0"
	ff = "/sys/class/net/%s/statistics/tx_bytes"%interface
	tx = open(ff).read().split('\n')[0] if os.path.exists(ff) else "0"

	if ipv6_native:
		# indy native IPv6
		ping6_native = test_ping6('2001:48d0:101:501:862b:2bff:fe17:497e')
	if ipv6_teredo:
		# indy Teredo server
		ping6_teredo = test_ping6('2001:0:c0ac:e290:5c:47cc:3f53:1d79'   )

	return (ipv4,ipv6_native,ipv6_teredo,rx,tx,ping6_native,ping6_teredo)

def network_bsd(interface):

	(ipv4,ipv6_native,ipv6_teredo,rx,tx,ping6_native,ping6_teredo) = ('','','','0','0','2','2')

	# This is iffy. The FreeBSD boxes use different interface names.
	# I found four so far. Loop over all of them and stop when the first
	# one returns something, unless the interface is explicitly specified.

	if interface:
		ipv4, ipv6_native, rx, tx, ping6_native = interface_bsd( interface )
	else:
		# ams-nl: em0
		# fnl-us: fxp0
		for interface in ('bce0','bge0','em0','vr0','fxp0'):
			ipv4, ipv6_native, rx, tx, ping6_native = interface_bsd( interface )
			if ipv4 != '':
				break

	return (ipv4,ipv6_native,ipv6_teredo,rx,tx,ping6_native,ping6_teredo)

def interface_bsd( interface ):

	# Check ifconfig for presence of network interface.
	# If it exists, extract native IPv4 and IPv6 addresses.

	(ipv4,ipv6_native,rx,tx,ping6_native) = ('','','0','0','2')

	ifconfig = os.popen('ifconfig %s 2> /dev/null'%interface).read()
	if len(ifconfig) == 0:
		return (ipv4,ipv6_native,rx,tx,ping6_native)

	# Native IPv4 address. Could be a public IP addresses or
	# (for NATted node) a private address like 192.168.*.* or 10.*.*.*.
	# inet addr:127.0.0.1  Mask:255.0.0.0
	# inet addr:10.1.0.140  Bcast:10.1.0.255  Mask:255.255.255.0

	# inet 195.148.124.66 netmask 0xffffffc0 broadcast 195.148.124.127 
	m = search('\s+inet\s([0-9,.]+)\s+netmask',ifconfig,MULTILINE)
	if m:
		ipv4 = m.group(1)

	# Native IPv6 address.
	# inet6 addr: ::1/128 Scope:Host
	# inet6 addr: 2001:0:c0ac:e290:5c:47cc:3f53:1d79/32 Scope:Global
	# inet6 2001:4b20:ca1d::e4 prefixlen 48 (zrh2-ch)

	m = search('\s+inet6\s(.*)\sprefixlen\s(64|48)\s(autoconf|$)',ifconfig,MULTILINE)
	if m:
		ipv6_native = m.group(1)

	# Check whether IPv6 responds to ping6.
	# Use specified interface, send one packet, wait 3 seconds
	# Indy: IPv6  : 2001:48d0:101:501:862b:2bff:fe17:497e
	#       Teredo: 2001:0:c0ac:e290:5c:47cc:3f53:1d79

	if ipv6_native:
		ping6_native = test_ping6('2001:48d0:101:501:862b:2bff:fe17:497e')

	# Bytes received and send.
	# RX bytes:0 (0.0 B)  TX bytes:0 (0.0 B)

	m = search('RX\sbytes:(\d+).*TX\sbytes:(\d+)',ifconfig,MULTILINE)
	if m:
		(rx,tx) = (m.group(1),m.group(2))

	return (ipv4,ipv6_native,rx,tx,ping6_native)

def monitor_uptime():		# uptime in seconds

	path = '/proc/uptime'
	if os.path.isfile(path):	# Should work on Raspberry Pi
		uptime = open(path,'r').read().split(' ')[0]
	else:						# Fall back for FreeBSD
		# kern.boottime: { sec = 1446442181, usec = 141832 } Mon Nov  2 05:29:41 2015
		uptime = os.popen('sysctl kern.boottime').read()
		uptime = search('sec\s=\s(\d+),',uptime)
		uptime = '%s'%(time()-int(uptime.group(1))) if uptime else '-1'

	return uptime

def app_status(prog,folder):

	# Do not return status = 0 or 1
	# 0: app not in use, or don't want to know)
	# 1: app is in use, but is not in the process table
	#    (i.e. it is not running at all)

	status = '2'		# Assume all is well

	if prog == 'teamx-driver':

		folder = os.path.join(folder,'team-probing')

		# Scamper is restarted for each new target list, so the timestamp
		# on the scamper.pid file must be recent. The scamper instance with
		# the pid in scamper.pid does not need to be running (it won't be
		# from the time one target list completes, and the next starts).

		pid_file = os.path.join(folder, 'scamper.pid')
		if os.path.exists( pid_file ) and time()-os.stat( pid_file ).st_mtime > 3600*6:
			status = '3'			# Team-probing is stuck

	elif prog == 'v6-prober':

		folder = os.path.join(folder,'topo-v6')

		# topo-v6 runs a scamper daemon. The pid shouldn't change.
		# Check for the scamper instance; if it is not running, bail
		# NOTE: I have seen cases where a topo-v6 scamper was running
		# with a pid that did not match the pid in scamper.pid, so
		# I'm not sure this is a valid test.

		pid_file = os.path.join(folder, 'scamper.pid')
		if os.path.exists( pid_file ):
			pid = int( open( pid_file ).read() )
			if not psutil.pid_exists( pid ):
				status = '3'			# Topo-v6 is stuck

		if status == '2':
			# Is there a way to check for an existing process that hasn't done anything
			# for a while (no cpu cycles, no i/o)

			# Stinks but try anyway: check for old inprogress file.

			# If it is too old, bail
			# This is not foolproof: the inprogress file may not exist at all
			# (this happens when the prober is waiting for the next list of targets).
			# It also helps if leftover inprogress files from aborted runs are
			# cleaned up.

			data_file = ''
			for f in os.listdir( folder ):
				if '.inprogress' in os.path.split(f)[1]:
					data_file = f if data_file == '' else f if f > data_file else data_file

			# The .in_progress file is added to as traceroutes complete
			# (maybe delayed by a delayed write). Modification time should be
			# pretty recent; let's say less than an hour old?

			# Not fool proof: if the process gets stuck before the .inprogress
			# is created, then this will be missed (and stuck=3 is returned).

			# More importantly, this could also mean that IPv6 simply has stopped
			# working, and that no traceroutes are added anymore (in which
			# case IPv6 probing should be turned off).

			if data_file:
				data_file = os.path.join( folder, data_file )
				file_stat = os.stat( data_file )
				if time()-file_stat.st_mtime > 3600*6:				# More than 6 hours since last write
					status = '4' if file_stat.st_size > 0 else '5'	# Check file size
			else:
				status = '6'

	# IPv4
	# 2: not stuck
	# 3: stuck

	# IPv6
	# 2: not stuck
	# 3: no topo-v6 scamper daemon (stuck? but maybe just pid mismatch)
	# 4: old .inprogress file with non-zero size (stuck? but could simply be a leftover file)
	# 5: old .inprogress file with size zero (ipv6 not working?)
	# 6: no .inprogress file (no v6 probing, or just completed target list)

	return status

def whatis_running(folder,hostname):

	# The 'n'-fields need to match up with ark-ping, so don't change them if possible.
	# Adding to the end is fine. If the numbering does get changed, then delete the old
	# data from the PingHome state-db before activating the change.

	apps = loads( open(ark_ping_apps).read() ) \
		if os.path.exists(ark_ping_apps) else {}

	activities = open(ark_activity_config).read().split('\n')[0] \
		if os.path.exists(ark_activity_config) else 'ARK_ACTIVITY_LIST='

	uname = os.uname()[0]
	for name in apps:
		in_use = apps[name]['in-use']
		apps[name]['in-use'] = uname == 'Linux' if in_use == 'os' else True if in_use == 'true' else in_use in activities

	names  = ('bdr','tsp','loss','wake')
	files  = dict( list(zip( names, [ os.path.join(ark_dir,'cron-entries','60-start-%s.crontab'%name) for name in names ] )) )
	states = dict( list(zip( names, [ os.path.join(activity_folder,'tsp-indir',   '%s.state'%name) for name in names ] )) )

	all_names = sorted( list(apps.keys()), key=lambda name: apps[name]['n'] )

	# On fnl-us (FreeBSD 12) the psutil package doesn't work, so removed it.

	try:
		global psutil
		import psutil
	except:
		all_status = dict( list(zip(all_names, ['0']*len(all_names))) )
	else:
		# Set status to 0 (not used) or 1 (used)

		all_status = dict( list(zip(all_names, ['0' if not apps[name]['in-use'] else '1' for name in all_names])) )

		# Congestion

		for name in files:
			if name in all_status and all_status[name] == '1' and os.path.isfile(files[name]):
				all_status[name] = '2' if os.path.isfile(states[name]) and os.path.getsize(states[name]) > 0 else '3'

		# Other Ark processes

		from inspect import ismethod

		for proc in psutil.process_iter():
			try:		 # Seen this crash on a zombie process.
				name = proc.name() if ismethod(proc.name) else proc.name
			except:
				continue

			#if prog == 'tmux' or prog == 'tmux:':
			#	pid = proc.pid() if ismethod(proc.pid) else proc.pid
			#	if os.path.exists("/proc/%d"%pid):
			#		uid = os.stat("/proc/%d"%pid).st_uid
			#		username = pwd.getpwuid(uid)[0]
			#		if username == 'ark':
			#			all_status['tmux'] = '2'	# In utils
			#		continue

			if name == 'ruby':
				cmd = proc.cmdline() if ismethod(proc.cmdline) else proc.cmdline
				if type(cmd) is list:
					cmd = ' '.join(cmd)
				for name in [ name for name in all_names if apps[name]['type'] == 'ruby' ]:
					if all_status[name] == '1' and name in cmd:
						all_status[name] = app_status(name,folder)
			elif name in apps and apps[name]['type'] == 'util' and all_status[name] == '1':
				all_status[name] = '2'

	# 0: no psutil or not in use
	# 1: not running
	# 2: running
	# 3: running, but stuck or idle? Manually kill and restart?

	return ''.join( [ all_status[name] for name in all_names ] )

def ntpd_usage():

	# Pick up the ntp server from the ntpd config file

	if not os.path.exists('/etc/ntp.conf'):
		return '0|0||'

	cf = open('/etc/ntp.conf','r').read()
	if not cf:
		return '0|0||'

	m = search('^server\s+(.*?)\s',cf,MULTILINE)
	if not m:
		return '0|0||'

	server_cf = m.group(1)		# ntp server in ntp.conf
	offset    = 0.0
	ok        = 0
	server    = ''

	# *141.82.25.201   131.188.3.220    2 u    3   64  377    0.542   -0.007   0.012
	cmd = os.popen('ntpq -pn 2> /dev/null').read()
	if len(cmd) > 0:

		# Really should be checking here for the server used for syncing (line starting with *)
		# Also should be checking for IPv6 addresses to catch IPv6 time servers
		# Detects both IPv4 and IPv6? Doesn't work in Python?
		#'/^(?>(?>([a-f0-9]{1,4})(?>:(?1)){7}|(?!(?:.*[a-f0-9](?>:|$)){8,})((?1)(?>:(?1)){0,6})?::(?2)?)|(?>(?>(?1)(?>:(?1)){5}:|(?!(?:.*[a-f0-9]:){6,})(?3)?::(?>((?1)(?>:(?1)){0,4}):)?)?(25[0-5]|2[0-4][0-9]|1[0-9]{2}|[1-9]?[0-9])(?>\.(?4)){3}))$/iD'

		m = search('\*(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}).*\d+.+\d+\s+\d+\s+(\d+)\s+\S+\s+(\S)\s+',cmd)
		if not m:
			m = search('(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}).*\d+.+\d+\s+\d+\s+(\d+)\s+\S+\s+(\S+)\s+',cmd)
		if m:
			server = m.group(1)			# ntp server used
			reach  = int(m.group(2))
			offset = float(m.group(3))
			if reach > 0:
				ok = reach

	offset = '%s'%(int(round(1000*offset)))

	return '%s|%s|%s|%s'%(offset,ok,server,server_cf)

def dns_resolve():
	result = '0|'

	# Pick up the ntp server from the ntpd config file

	if not os.path.exists('/etc/resolv.conf'):
		return result

	cf = open('/etc/resolv.conf','r').read()
	if not cf:
		return result

	m = search('^nameserver\s+(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})',cf,MULTILINE)
	if not m:
		return result

	server = m.group(1)

	ok = call(['ping','-c','1','-W','3','indy.caida.org'],stdout=open('/dev/null','w'),stderr=open('/dev/null','w'))
	ok = 1 if ok == 0 else 0
	ok = '%s'%ok

	return '%s|%s'%(ok,server)

def hardware():

	if os.uname()[0] == 'Linux':
		m = search('(?:PRETTY_NAME|VERSION)="(.+)"', open('/etc/os-release','r').read())
		pretty_name = m.group(1) if m else ''

		cpu_hardware	= ""
		cpu_revision	= ""
		cpu_serial		= ""
		hw_model		= ""

		# RPi
		m = search('^Hardware\s+:\s(.+)\nRevision\s+:\s(.+)\nSerial\s+:\s(.+)',open('/proc/cpuinfo','r').read(),MULTILINE)
		if m:
			cpu_hardware	= m.group(1)
			cpu_revision	= m.group(2)
			cpu_serial		= m.group(3)
			hw_model		= open('/proc/device-tree/model').read()[0:-1]

		else:
			m = search('^vendor_id\s+:\s(.+)\ncpu\sfamily\s+:\s(.+)\nmodel\s+:\s.+\nmodel\sname\s+:\s(.+)',open('/proc/cpuinfo','r').read(),MULTILINE)
			if m:			# Ubuntu
				cpu_hardware	= m.group(1)
				cpu_revision	= m.group(2)
				hw_model		= ' '.join(m.group(3).split())

		cpu_arch = os.popen('lscpu | grep Architecture'  ).read().split(' ')[-1][0:-1]
		memory   = os.popen('grep MemTotal /proc/meminfo').read().split('\n')[0].split()[-2]

		interface = interface_name()

		ff = "/sys/class/net/%s/address"%interface
		mac = open(ff).read()[0:-1] if os.path.exists(ff) else ""

		#pphick@san-us:~$ netstat -rn
		#Kernel IP routing table
		#Destination     Gateway         Genmask         Flags   MSS Window  irtt Iface
		#0.0.0.0         192.172.226.1   0.0.0.0         UG        0 0          0 eth0
		#192.172.226.0   0.0.0.0         255.255.255.0   U         0 0          0 eth0

		gw_address = os.popen('netstat -rn 2> /dev/null').read().split('\n')[0:-1]
		if len(gw_address) > 2:
			gw_address = gw_address[2].split()[1]
			gw_mac = os.popen('arp -n %s 2> /dev/null'%gw_address).read()[0:-1].split('\n')[1].split()[2]
		else:
			gw_address = "0.0.0.0"
			gw_mac = ""

		#pphick@san-us:~$ arp -n 192.172.226.1
		#Address                  HWtype  HWaddress           Flags Mask            Iface
		#192.172.226.1            ether   00:00:5e:00:01:ea   C                     eth0

		# image_id = open( '/usr/local/ark/config/image_id.txt', 'w' ).read().split('\n')[0]
		# uname = 'Linux 6.1.19-v8+'

		uname  = ' '.join(os.uname()[slice(0,3,2)])	# sysname + release (kernel version)
		result = '%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s'%(
			pretty_name		, # OS name
			uname			, # OS kernel
			hw_model		, # Hardware model
			mac				, # Interface mac
			memory			, # Memory
			cpu_hardware	, # CPU hardware
			cpu_revision	, # CPU revision
			cpu_serial		, # CPU serial
			cpu_arch		, # armv6l or armv7l
			gw_address		, # Gateway address
			gw_mac			, # Gateway mac
		)

	else:
		#(zrh2-ch) ~$ netstat -rn
		#Routing tables
		#
		#Internet:
		#Destination        Gateway            Flags      Netif Expire
		#default            81.94.127.241      UGS         em0
		#...

		gw_address = os.popen('netstat -rn 2> /dev/null').read().split('\n')[0:-1][4].split()[1]

		#(zrh2-ch) ~$ arp -n 81.94.127.241
		#? (81.94.127.241) at 00:12:f2:9b:2b:00 on em0 expires in 1200 seconds [ethernet]

		# uname = 'FreeBSD 10.1-RELEASE-p9'
		uname  = ' '.join(os.uname()[slice(0,3,2)])	# sysname + release (kernel version)

		result = '%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s'%(
			uname.split('-')[0]								,	# OS kernel
			uname											,	# OS name
			' '.join(os.popen('sysctl hw.model').read()[0:-1].split()[1:]),	# Model
			''												,	# Mac
			int(os.popen('sysctl hw.physmem').read().split('\n')[0].split()[-1])/1000, # Memory
			''												,	# CPU hardware
			''												,	# CPU revision
			''												,	# CPU serial
			os.popen('sysctl hw.machine').read()[0:-1].split()[1],	# i386
			gw_address,											# Gateway address
			os.popen('arp -n %s 2> /dev/null'%gw_address).read()[0:-1].split('\n')[0].split()[3],	# Gateway mac
		)

	return result

def storage_device():
	# https://www.cameramemoryspeed.com/sd-memory-card-faq/reading-sd-card-cid-serial-psn-internal-numbers/

	ff = "/proc/device-tree/model"
	if os.path.exists(ff) and 'Raspberry' in open(ff).read().split('\x00')[0]:

		sd_manfid	= open('/sys/block/mmcblk0/device/manfid').read().split('\n')[0]
		sd_model	= open('/sys/block/mmcblk0/device/name'  ).read().split('\n')[0]
		sd_serial	= open('/sys/block/mmcblk0/device/serial').read().split('\n')[0]
		sd_date		= open('/sys/block/mmcblk0/device/date'  ).read().split('\n')[0]

		sd_size		= str(int(os.popen('blockdev --getsz /dev/mmcblk0').read().split('\n')[0])//2)

		blk0		= os.popen('mount | grep /dev/mmcblk0').read().split('\n')[0:-1]
		blk0.sort()

		sd_state = []
		for p in blk0:
			m = re.search( '^/dev/mmcblk0p\d\s.*\((.*?),.*\)$', p )
			if m:
				sd_state.append( m.group(1) )

	else:
		sd_manfid	= ''
		sd_model	= ''
		sd_serial	= ''
		sd_date		= ''

		sd_size		= os.popen('df -k /').read().split('\n')[1].split()[1]

		sd_state = [ '', '' ]

	result = '%s|%s|%s|%s|%s|%s|%s'%(
		sd_manfid	, # SD card manufacturer ID
		sd_model	,
		sd_serial	, # SD card serial number
		sd_date		,

		sd_size		, # SD card size in 1K blocks

		sd_state[0]	,
		sd_state[1]	,
	)

	return result

def whatis_activity(folder):

	sizes = '0/0/0|'
	if not os.path.isdir(folder):		# Not a directory
		return sizes
	if len(os.listdir(folder)) == 0:	# No subdirectories
		return sizes

	sizes = os.popen('df -k %s'%folder).read().split('\n')[1].split()
	sizes = sizes[3]+'/'+sizes[2]+'/'+sizes[1] # Available/available/Total/

	sep = '|'
	disk_usage = os.popen('du -d 0 %s/*'%folder).read().split('\n')[0:-1]
	for entry in disk_usage:			# Should be at least one subdirectory
		s,f = entry.split('\t')
		if os.path.isdir(f):
			f = os.path.split(f)[1]
			sizes += '%s%s:%s'%(sep,f,s)# |name:size
			sep = ','

	sizes += '|'
	if os.path.exists(ark_activity_config):
		m = search('ARK_ACTIVITY_LIST=(.*)', open(ark_activity_config,'r').read())
		if m:
			sizes += m.group(1)

	#(xxx-yy|3|735376/2270160/3121684|bulk-pinger:512,...,yarrp6:20|team-probing,topo-on-demand)
	return sizes

if __name__ == "__main__":

	from optparse import OptionParser

	# Pick up the hostname. Usually an Ark node, but could be a CAIDA machine.
	# First choice is ARK_MONITOR in the monitor-config file. This should work
	# on every Ark node. Fallback is the network hostname (in case the script is run
	# from a CAIDA machine).

	hostname = None			# Default for options.node
	is_ark_node = False
	if os.path.exists(ark_monitor_config):
		is_ark_node = True	# Is monitor-config exists, it is an Ark node. Is this good enough?
		m = search('ARK_MONITOR=(.+)', open(ark_monitor_config,'r').read())
		if m:
			hostname = m.group(1)
	if not hostname:
		hostname = os.popen('hostname -s').read()[0:-1]

		# Some FreeBSD boxes have deviant hostnames (only fnl-us??)
		# (should never happen because monitor-config file exists)

		#if hostname == 'ark':
		#	hostname = os.popen('hostname').read()[0:-1]
		#	if hostname == 'ark.caida.netsec.colostate.edu':
		#		hostname = 'fnl-us' 

	version = '2.05'
	usage = "%prog [--delay|jitter <n>] <address>\n\taddress: indy|malibu|irori|a.b.c.d (default is indy)"

	parser = OptionParser(usage=usage,version=version)

	parser.add_option('-v', '--verbose'	,
		dest		= 'verbose'			,
		action		= 'store_true'		,
		default		= False				,
		help		= 'verbose mode'	,
	)

	parser.add_option('-n', '--dry-run'	,
		dest		= 'dryrun'			,
		action		= 'store_true'		,
		default		= False				,
		help		= 'make dryrun'		,
	)

	parser.add_option('-j', '--jitter'	,
		dest		= 'jitter'			,
		action		= 'store'			,
		type		= 'int'				,
		default		= 120				,
		help		= 'maximum random time delay (in sec) before ping'
	)

	parser.add_option('-d', '--delay'	,
		dest		= 'delay'			,
		action		= 'store'			,
		type		= 'int'				,
		default		= 0					,
		help		= 'fixed time delay (in sec) before ping'
	)

	parser.add_option('-b', '--boot'	,
		dest		= 'boot'			,
		action		= 'store_true'		,
		default		= False 			,
		help		= 'mark packet as sent at boot'	,
	)

	parser.add_option('-x', '--network-traffic'	,
		dest		= 'network_traffic'			,
		action		= 'store_true'				,
		default		= False						,
		help		= 'send TX and RX bytes'	,
	)

	parser.add_option('', '--ntpd-usage'		,
		dest		= 'ntpd_usage'				,
		action		= 'store_true'				,
		default		= False 					,
		help		= 'send ntpd usage'			,
	)

	parser.add_option('', '--dns-resolve'		,
		dest		= 'dns_resolve'				,
		action		= 'store_true'				,
		default		= False 					,
		help		= 'send dns server'			,
	)

	parser.add_option('-a', '--activity-usage'	,
		dest		= 'activity_usage'			,
		action		= 'store_true'				,
		default		= False 					,
		help		= 'send activity usage'		,
	)

	parser.add_option('', '--hardware'			,
		dest		= 'hardware'				,
		action		= 'store_true'				,
		default		= False 					,
		help		= 'send hardware parameters',
	)

	parser.add_option('', '--storage-device'	,
		dest		= 'storage_device'			,
		action		= 'store_true'				,
		default		= False 					,
		help		= 'send SD card/hard drive parameters',
	)

	parser.add_option('-i', '--interface'		,
		dest		= 'interface'				,
		action		= 'store'					,
		help		= 'interface to extract IP address from (FreeBSD only)'	,
	)

	parser.add_option('', '--encrypt'			,
		dest		= 'encrypt'					,
		action		= 'store'					,
		default		= ''						,
		help		= 'public key file'			,
	)

	parser.add_option('-t', '--to'		,
		dest		= 'to'				,
		action		= 'store'			,
		default		= 'indy'			,
		help		= 'host destination of ICMP packet',
	)

	parser.add_option('', '--node'		,
		dest		= 'node'			,
		action		= 'store'			,
		default		= hostname			,
		help		= 'Ark node to be used for applying data',
	)

	options, args = parser.parse_args()

	caida_ips = '192.172.226.'			# CAIDA /24
	destination_ips = {
		'indy'	: caida_ips+'134',
		'malibu': caida_ips+'106',
		'irori' : caida_ips+'23' ,
		'nabu'  : caida_ips+'24' ,
	}

	# activity_folder = os.path.join(ark_dir, 'activity')

	# Sleep to allow for network to start up (needed for boot packet)

	if options.delay > 0:
		if options.verbose:
			print("sleeping %d secs"%options.delay)
		sleep( options.delay )

	if not options.dryrun:
		delay = 0 if options.delay > 0 else options.jitter*random() if options.jitter > 0 else 0
		if delay > 0:
			if options.verbose:
				print("sleeping %d secs before ping"%delay)
			sleep( delay )

	# Packet is send to indy by default.

	destination_ip = destination_ips[options.to] if options.to in destination_ips else options.to

	ipv4,ipv6_native,ipv6_teredo,rx,tx,ping6_native,ping6_teredo =					\
			('','','','0','0','2','2')		if not is_ark_node				else	\
			network_bsd(options.interface)	if os.uname()[0] == 'FreeBSD'	else	\
			network_rpi()

	if len(args) > 0:

		data = {}
		for arg in args:
			p = arg.find('=')
			if p < 1:
				print('invalid key/value pair, %s'%arg)
				continue
			key = arg[0:p]
			val = arg[p+1:]
			if val[0] == "'" or val[0] == '"':
				if val[0] == val[-1]:
					val = val[1:-1]
			data[key] = val

		# The keys are not validated here. That is done by ark-ping-watch.py

		data = '|'.join( [ k+'='+data[k] for k in data ] )
		data =	'('+ options.node			+ \
				'|99'						+ \
				'|%s'%data					+ \
		')'

	elif not is_ark_node:

		# This is send if the monitor-config does not exist yet.

		# This still is the same as the hourly heartbeat packet.
		# It probably should get a separate ID.

		data =	'('+ hostname				+ \
				'|0'						+ \
				'|'+ ipv4					+ \
				'|'+ ipv6_teredo			+ \
				'|'+ ping6_teredo			+ \
				'|'+ ipv6_native			+ \
				'|'+ ping6_native			+ \
				'|'+ monitor_uptime()		+ \
				'|'+ whatis_running(activity_folder,hostname)		+ \
				')'

	# Boot packet: should be set up to run in cron at boot.
	# It contains addresses only so that address changes can be processed
	# instantly.

	elif options.boot:

		interface = interface_name()
		ff = "/sys/class/net/%s/address"%interface
		mac = open(ff).read()[0:-1] if os.path.exists(ff) else ""

		uname = ' '.join( os.uname()[slice(0,3,2)] )
		data =	'('+ hostname				+ \
				'|1'						+ \
				'|'+ ipv4					+ \
				'|'+ ipv6_teredo			+ \
				'|'+ ping6_teredo			+ \
				'|'+ ipv6_native			+ \
				'|'+ ping6_native			+ \
				'|'+ uname					+ \
				'|'+ mac					+ \
				')'

	# Network traffic (once a day?)

	elif options.network_traffic:

		data =	'('+ hostname				+ \
				'|2'						+ \
				'|'+ rx						+ \
				'|'+ tx						+ \
				'|'+ vnstat_rate()			+ \
				')'

	# Disk usage (once a day?)

	elif options.activity_usage:
		data =	'('+ hostname				+ \
				'|3'						+ \
				'|'+ whatis_activity(activity_folder)	+ \
				')'

	# Ntp usage (once a day?)

	elif options.ntpd_usage:
		data =	'('+ hostname				+ \
				'|4'						+ \
				'|'+ ntpd_usage()			+ \
				')'

	elif options.dns_resolve:
		data =	'('+ hostname				+ \
				'|5'						+ \
				'|'+ dns_resolve()			+ \
				')'

	elif options.hardware:
		data =	'('+ hostname				+ \
				'|6'						+ \
				'|'+ hardware()				+ \
				')'

	elif options.storage_device:
		data =	'('+ hostname				+ \
				'|7'						+ \
				'|'+ storage_device()		+ \
				')'

	# Default hourly packet.

	else:
		data =	'('+ hostname				+ \
				'|0'						+ \
				'|'+ ipv4					+ \
				'|'+ ipv6_teredo			+ \
				'|'+ ping6_teredo			+ \
				'|'+ ipv6_native			+ \
				'|'+ ping6_native			+ \
				'|'+ monitor_uptime()		+ \
				'|'+ whatis_running(activity_folder,hostname)		+ \
				')'

	if options.dryrun:
		print("%s %s\nlength: %s/192"%(destination_ip, data, len(data)))

	else:

		#if len(data) < 192:
		#	data += (192-len(data))*'.'

		if options.encrypt:
			from pickle				import load
			from Crypto.PublicKey	import RSA
			public_key_file = open( options.encrypt, 'r')
			public_key = load(public_key_file)
			public_key_file.close()
			data = public_key.encrypt(data,32)

		# The source address of the packet is either the IPv4 address of the
		# Ark node itself, or (if the node is behind a NAT) the public address
		# of the router.

		if options.verbose:
			print("%s %s\nlength: %s/192"%(destination_ip, data, len(data)))
			# verbosely send echo-request
			verbose(destination_ip, 1, 4, data=data.encode("ascii"))
		else:
			# quietly send echo-request
			echo(destination_ip, 4, data=data.encode("ascii"))

	exit()
